(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/drivers/mail.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
var _this = this;                                                      //
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var Future = Npm.require("fibers/future");                             // 18
var Promise = Npm.require("es6-promise").Promise;                      // 19
var Capnp = Npm.require("capnp");                                      // 20
                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");            // 22
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;
var EmailSendPort = EmailRpc.EmailSendPort;                            // 25
                                                                       //
var Url = Npm.require("url");                                          // 27
                                                                       //
var ROOT_URL = Url.parse(process.env.ROOT_URL);                        // 29
var HOSTNAME = ROOT_URL.hostname;                                      // 30
                                                                       //
var DAILY_LIMIT = 50;                                                  // 32
var RECIPIENT_LIMIT = 20;                                              // 33
                                                                       //
var CLIENT_TIMEOUT = 15000; // 15s                                     // 35
                                                                       //
// Every day, reset all per-user sent counts to zero.                  //
// TODO(cleanup): Consider a more granular approach. For example, each user could have a timer
//   after which their count will reset. We'd only check the timer when that user is trying to
//   send a new message. This avoids a global query.                   //
if (!Meteor.settings.replicaNumber) {                                  // 41
  // only first replica                                                //
  SandstormDb.periodicCleanup(86400000, function () {                  // 42
    Meteor.users.update({ dailySentMailCount: { $exists: true } }, { $unset: { dailySentMailCount: "" } }, { multi: true });
  });                                                                  //
}                                                                      //
                                                                       //
Meteor.startup(function () {                                           // 49
  var SANDSTORM_SMTP_PORT = parseInt(process.env.SANDSTORM_SMTP_PORT, 10) || 30025;
  var BIND_IP = process.env.BIND_IP || "127.0.0.1";                    // 51
                                                                       //
  simplesmtp.createSimpleServer({ SMTPBanner: "Sandstorm Mail Server" }, function (req) {
    var mailparser = new MailParser();                                 // 54
    req.pipe(mailparser);                                              // 55
                                                                       //
    mailparser.on("end", function (mail) {                             // 57
      // Wrap in outer promise for easier error handling.              //
      Promise.resolve().then(function () {                             // 59
        // Extract the 'from' address.                                 //
        var from = undefined;                                          // 61
        if (mail.from && mail.from.length > 0) {                       // 62
          // It's theoretically possible for the message to have multiple 'from' headers, but this
          // never really happens in legitimate practice so we'll just take the first one.
          from = mail.from[0];                                         // 65
        } else {                                                       //
          // The mail body is missing a 'From:' header. We'll use req.from instead. Note that
          // req.from is actually the bounce address, which is sometimes *not* the original sender
          // but rather some intermediate agent (e.g. a mailing list daemon). See:
          //   http://en.wikipedia.org/wiki/Bounce_address             //
          // TODO(someday): Is this really right, or should we report a blank address instead?
          from = { address: req.from, name: false };                   // 72
        }                                                              //
                                                                       //
        var attachments = [];                                          // 75
        if (mail.attachments) {                                        // 76
          attachments = mail.attachments.map(function (attachment) {   // 77
            var disposition = attachment.contentDisposition || "attachment";
            disposition += ';\n\tfilename="' + (attachment.fileName || attachment.generatedFileName) + '"';
            return {                                                   // 80
              contentType: attachment.contentType,                     // 81
              contentDisposition: disposition,                         // 82
              contentId: attachment.contentId,                         // 83
              content: attachment.content                              // 84
            };                                                         //
          });                                                          //
        }                                                              //
                                                                       //
        if (mail.replyTo && mail.replyTo.length > 1) {                 // 89
          console.error("More than one reply-to address address was received in an email.");
        }                                                              //
                                                                       //
        var mailMessage = {                                            // 93
          // Note that converting the date to nanoseconds actually goes outside the range of
          // integers that Javascript can represent precisely. But this is OK because dates aren't
          // precise anyway.                                           //
          date: (mail.date || new Date()).getTime() * 1000000,         // 97
          from: from,                                                  // 98
          to: mail.to,                                                 // 99
          cc: mail.cc || [],                                           // 100
          bcc: mail.bcc || [],                                         // 101
          replyTo: mail.replyTo && mail.replyTo[0] || {},              // 102
          messageId: mail.headers["message-id"] || Meteor.uuid() + "@" + HOSTNAME,
          references: mail.references || [],                           // 104
          inReplyTo: mail.inReplyTo || [],                             // 105
          subject: mail.subject || "",                                 // 106
          text: mail.text || null,                                     // 107
          html: mail.html || null,                                     // 108
          attachments: attachments                                     // 109
        };                                                             //
                                                                       //
        // Get list of grain IDs.                                      //
        var grainPublicIds = _.uniq(req.to.map(function (deliverTo) {  // 113
          // simplesmtp already validates that the address contains an @.
          // To simplify things, we ignore the hostname part of the address and assume that the
          // message would not have been sent here if it weren't intended for our host. Usually
          // there will be an nginx frontend verifying hostnames anyway. Grain public IDs are
          // globally unique anyway, so an e-mail meant for another server presumably won't match
          // any ID at this one anyway.                                //
          return deliverTo.slice(0, deliverTo.indexOf("@"));           // 120
        }));                                                           //
                                                                       //
        // Deliver to each grain in parallel.                          //
        return Promise.all(grainPublicIds.map(function (publicId) {    // 124
          // Wrap in a function so that we can call it recursively to retry.
          var tryDeliver = function (retryCount) {                     // 126
            var grainId = undefined;                                   // 127
            return inMeteor(function () {                              // 128
              var grain = Grains.findOne({ publicId: publicId }, { fields: {} });
              if (grain) {                                             // 130
                grainId = grain._id;                                   // 131
                return globalBackend.continueGrain(grainId, retryCount > 0);
              } else {                                                 //
                // TODO(someday): We really ought to rig things up so that the 'RCPT TO' SMTP command
                //   fails in this case, but simplesmtp doesn't appear to support that.
                throw new Error("No such grain: " + publicId);         // 136
              }                                                        //
            }).then(function (grainInfo) {                             //
              var supervisor = grainInfo.supervisor;                   // 139
              var uiView = supervisor.getMainView().view;              // 140
                                                                       //
              // Create an arbitrary struct to use as the session params. E-mail sessions actually
              // require no params, but node-capnp won't let us pass null and we don't have an
              // EmptyStruct type available, so we just use EmailAddress, but any struct type would
              // work.                                                 //
              // TODO(cleanup): Fix node-capnp to accept null.         //
              var emptyParams = Capnp.serialize(EmailRpc.EmailAddress, {});
                                                                       //
              // Create a new session of type HackEmailSession. This is a short-term hack until
              // persistent capabilities and the Powerbox are implemented. We need to pass along a
              // HackSessionContext because old versions of sandstorm-http-bridge always use
              // the context of the most recent session; if an old RoundCube grain has a
              // WebSession open, receives an email, and then tries to send an email, the request
              // will go out on the SessionContext associated with the HackEmailSession that
              // delivered the email.                                  //
              var session = uiView.newSession({}, makeHackSessionContext(grainId), "0xc3b5ced7344b04a6", emptyParams).session.castAs(EmailSendPort);
              return session.send(mailMessage);                        // 160
            })["catch"](function (err) {                               //
              if (SandstormBackend.shouldRestartGrain(err, retryCount)) {
                return tryDeliver(retryCount + 1);                     // 163
              } else {                                                 //
                throw err;                                             // 165
              }                                                        //
            });                                                        //
          };                                                           //
                                                                       //
          return tryDeliver(0);                                        // 170
        }));                                                           //
      }).then(function () {                                            //
        req.accept();                                                  // 173
      }, function (err) {                                              //
        console.error("E-mail delivery failure:", err.stack);          // 175
        req.reject(err.message);                                       // 176
      });                                                              //
    });                                                                //
  }).listen(SANDSTORM_SMTP_PORT, BIND_IP);                             //
});                                                                    //
                                                                       //
function formatAddress(field) {                                        // 182
  if (!field) {                                                        // 183
    return null;                                                       // 184
  }                                                                    //
                                                                       //
  if (Array.isArray(field)) {                                          // 187
    return field.map(formatAddress);                                   // 188
  }                                                                    //
                                                                       //
  if (field.name) {                                                    // 191
    return field.name + " <" + field.address + ">";                    // 192
  }                                                                    //
                                                                       //
  return field.address;                                                // 195
}                                                                      //
                                                                       //
hackSendEmail = function (session, email) {                            // 198
  return inMeteor((function () {                                       // 199
    var recipientCount = 0;                                            // 200
    recipientCount += email.to ? email.to.length : 0;                  // 201
    recipientCount += email.cc ? email.cc.length : 0;                  // 202
    recipientCount += email.bcc ? email.bcc.length : 0;                // 203
    if (recipientCount > RECIPIENT_LIMIT) {                            // 204
      throw new Error("Sorry, Sandstorm currently only allows you to send an e-mail to " + RECIPIENT_LIMIT + " recipients at a time, for spam control. Consider setting up a mailing list. " + "Please feel free to contact us if this is a problem for you.");
    }                                                                  //
                                                                       //
    // Overwrite the 'from' address with the grain's address.          //
    if (!email.from) {                                                 // 212
      email.from = {};                                                 // 213
    }                                                                  //
                                                                       //
    var grainAddress = session._getAddress();                          // 216
    var userAddress = session._getUserAddress();                       // 217
                                                                       //
    // First check if we're changing the from address, and if so, move it to reply-to
    if (email.from.address !== grainAddress && email.from.address !== userAddress.address) {
      throw new Error("FROM header in outgoing emails need to equal either " + grainAddress + " or " + userAddress.address + ". Yours was: " + email.from.address);
    }                                                                  //
                                                                       //
    var mc = new MailComposer();                                       // 226
                                                                       //
    mc.setMessageOption({                                              // 228
      from: formatAddress(email.from),                                 // 229
      to: formatAddress(email.to),                                     // 230
      cc: formatAddress(email.cc),                                     // 231
      bcc: formatAddress(email.bcc),                                   // 232
      replyTo: formatAddress(email.replyTo),                           // 233
      subject: email.subject,                                          // 234
      text: email.text,                                                // 235
      html: email.html                                                 // 236
    });                                                                //
                                                                       //
    var envelope = mc.getEnvelope();                                   // 239
    envelope.from = grainAddress;                                      // 240
                                                                       //
    mc.setMessageOption({                                              // 242
      envelope: envelope                                               // 243
    });                                                                //
                                                                       //
    var headers = {};                                                  // 246
    if (email.messageId) {                                             // 247
      mc.addHeader("message-id", email.messageId);                     // 248
    }                                                                  //
                                                                       //
    if (email.references) {                                            // 251
      mc.addHeader("references", email.references);                    // 252
    }                                                                  //
                                                                       //
    if (email.messageId) {                                             // 255
      mc.addHeader("in-reply-to", email.inReplyTo);                    // 256
    }                                                                  //
                                                                       //
    if (email.date) {                                                  // 259
      var date = new Date(email.date / 1000000);                       // 260
      if (!isNaN(date.getTime())) {                                    // 261
        // Check to make sure date is valid                            //
        mc.addHeader("date", date.toUTCString());                      // 262
      }                                                                //
    }                                                                  //
                                                                       //
    if (email.attachments) {                                           // 266
      email.attachments.forEach(function (attachment) {                // 267
        mc.addAttachment({                                             // 268
          cid: attachment.contentId,                                   // 269
          contentType: attachment.contentType,                         // 270
          contentDisposition: attachment.contentDisposition,           // 271
          contents: attachment.content                                 // 272
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    var grain = Grains.findOne(session.grainId);                       // 277
    if (!grain) throw new Error("Grain does not exist.");              // 278
                                                                       //
    var user = Meteor.users.findAndModify({                            // 280
      query: { _id: grain.userId },                                    // 281
      update: {                                                        // 282
        $inc: {                                                        // 283
          dailySentMailCount: 1                                        // 284
        }                                                              //
      },                                                               //
      fields: { dailySentMailCount: 1 }                                // 287
    });                                                                //
    if (user.dailySentMailCount >= DAILY_LIMIT) {                      // 289
      throw new Error("Sorry, you've reached your e-mail sending limit for today. Currently, Sandstorm " + "limits each user to " + DAILY_LIMIT + " e-mails per day for spam control reasons. " + "Please feel free to contact us if this is a problem.");
    }                                                                  //
                                                                       //
    SandstormEmail.rawSend(mc);                                        // 296
  }).bind(_this))["catch"](function (err) {                            //
    console.error("Error sending e-mail:", err.stack);                 // 298
    throw err;                                                         // 299
  });                                                                  //
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=mail.js.map
